"""Top-level package for Ganblr."""

__author__ = """Tulip Lab"""
__email__ = 'jhzhou@tuliplab.academy'
__version__ = '0.1.1'

from .kdb import KdbHighOrderFeatureEncoder
from .utils import get_demo_data

__all__ = ['models', 'KdbHighOrderFeatureEncoder', 'get_demo_data']


from .ganblr import GANBLR
from sklearn.mixture import BayesianGaussianMixture
from sklearn.preprocessing import MinMaxScaler, LabelEncoder, OrdinalEncoder
from scipy.stats import truncnorm
import numpy as np

class DMMDiscritizer:
    def __init__(self, random_state):
        self.__dmm_params = dict(weight_concentration_prior_type="dirichlet_process",
            n_components=2 * 5, reg_covar=0, init_params='random',
            max_iter=1500, mean_precision_prior=.1,
            random_state=random_state)

        self.__scaler = MinMaxScaler()
        self.__dmms = []
        self.__arr_mu = []
        self.__arr_sigma = []
        self._random_state = random_state

    def fit(self, x):
        """
        Do DMM Discritization.

        Parameter:
        ---------
        x (2d np.numpy): data to be discritize. Must bu numeric data.

        Return:
        ----------
        self

        """
        assert(isinstance(x, np.ndarray))
        assert(len(x.shape) == 2)

        x_scaled = self.__scaler.fit_transform(x)
        self.__internal_fit(x_scaled)
        return self

    def transform(self, x) -> np.ndarray:        
        x = self.__scaler.transform(x)
        arr_modes = []
        for i, dmm in enumerate(self.__dmms):
            modes = dmm.predict(x[:,i:i+1])
            modes = LabelEncoder().fit_transform(modes)#.astype(int)
            arr_modes.append(modes)
        return self.__internal_transform(x, arr_modes)

    def fit_transform(self, x) -> np.ndarray:
        assert(isinstance(x, np.ndarray))
        assert(len(x.shape) == 2)

        x_scaled = self.__scaler.fit_transform(x)
        arr_modes = self.__internal_fit(x_scaled)
        return self.__internal_transform(x_scaled, arr_modes)

    def __internal_fit(self, x):
        self.__dmms.clear()
        self.__arr_mu.clear()
        self.__arr_sigma.clear()

        arr_mode = []
        for i in range(x.shape[1]):
            cur_column = x[:,i:i+1]
            dmm = BayesianGaussianMixture(**self.__dmm_params)
            y = dmm.fit_predict(cur_column)
            lbe = LabelEncoder().fit(y)
            mu  = dmm.means_[:len(lbe.classes_)]
            sigma = np.sqrt(dmm.covariances_[:len(lbe.classes_)])

            arr_mode.append(lbe.transform(y))#.astype(int))
            #self.__arr_lbes.append(lbe)
            self.__dmms.append(dmm)
            self.__arr_mu.append(mu.ravel())
            self.__arr_sigma.append(sigma.ravel())
        return arr_mode

    def __internal_transform(self, x, arr_modes):
        _and = np.logical_and
        _not = np.logical_not

        discretized_data = []
        for i, (modes, mu, sigma) in enumerate(zip(
            arr_modes,
            self.__arr_mu, 
            self.__arr_sigma)):

            cur_column = x[:,i]
            cur_mu     = mu[modes]
            cur_sigma  = sigma[modes]
            x_std      = cur_column - cur_mu

            less_than_n3sigma = (x_std <= -3*cur_sigma)
            less_than_n2sigma = (x_std <= -2*cur_sigma)
            less_than_n1sigma = (x_std <=   -cur_sigma)
            less_than_0       = (x_std <=            0)
            less_than_1sigma  = (x_std <=    cur_sigma)
            less_than_2sigma  = (x_std <=  2*cur_sigma)
            less_than_3sigma  = (x_std <=  3*cur_sigma)
            
            discretized_x = 8 * modes
            discretized_x[_and(_not(less_than_n3sigma), less_than_n2sigma)] += 1
            discretized_x[_and(_not(less_than_n2sigma), less_than_n1sigma)] += 2
            discretized_x[_and(_not(less_than_n1sigma), less_than_0)]       += 3
            discretized_x[_and(_not(less_than_0)      , less_than_1sigma)]  += 4
            discretized_x[_and(_not(less_than_1sigma) , less_than_2sigma)]  += 5
            discretized_x[_and(_not(less_than_2sigma) , less_than_3sigma)]  += 6
            discretized_x[_not(less_than_3sigma)]                           += 7
            discretized_data.append(discretized_x.reshape(-1,1))
        
        return np.hstack(discretized_data)

    def inverse_transform(self, x, verbose=1) -> np.ndarray:
        x_modes = x // 8
        x_bins = x % 8
        
        def __sample_one_column(i, mu, sigma):
            cur_column_modes = x_modes[:,i]
            cur_column_bins  = x_bins[:,i]
            cur_column_mode_uniques    = np.unique(cur_column_modes)
            inversed_x = np.zeros_like(cur_column_modes, dtype=float)      

            for mode in cur_column_mode_uniques:
                cur_mode_idx = cur_column_modes == mode
                cur_mode_mu = mu[mode]
                cur_mode_sigma = sigma[mode]

                sample_results = self.__sample_from_truncnorm(cur_column_bins[cur_mode_idx], cur_mode_mu, cur_mode_sigma, random_state=self._random_state)
                inversed_x[cur_mode_idx] = sample_results

            return inversed_x.reshape(-1,1)
        
        if verbose:
            from tqdm import tqdm
            _progress_wrapper = lambda iterable: tqdm(iterable, desc='sampling', total=len(self.__arr_mu))
        else:
            _progress_wrapper = lambda iterable: iterable
        inversed_data = np.hstack([__sample_one_column(i, mu, sigma)
            for i, (mu, sigma) in _progress_wrapper(enumerate(zip(self.__arr_mu, self.__arr_sigma)))])
        
        #the sampling progress is fast enough so there is no need for parallelization
        #inversed_data = np.hstack(Parallel(n_jobs=n_jobs, verbose=verbose)(delayed(__sample_one_column)(i, mu, sigma)
        #    for i, (mu, sigma) in enumerate(zip(self.__arr_mu, self.__arr_sigma))))
        return self.__scaler.inverse_transform(inversed_data)

    @staticmethod
    def __sample_from_truncnorm(bins, mu, sigma, random_state=None): 
        sampled_results = np.zeros_like(bins, dtype=float)
        def __sampling(idx, range_min, range_max):
            sampling_size = np.sum(idx)
            if sampling_size != 0:
                sampled_results[idx] = truncnorm.rvs(range_min, range_max, loc=mu, scale=sigma, size=sampling_size, random_state=random_state)
        
        #shape param (min, max) of scipy.stats.truncnorm.rvs are still defined with respect to the standard normal
        __sampling(bins == 0, np.NINF, -3)
        __sampling(bins == 1, -3, -2)
        __sampling(bins == 2, -2, -1)
        __sampling(bins == 3, -1,  0)
        __sampling(bins == 4, 0,   1)
        __sampling(bins == 5, 1,   2)
        __sampling(bins == 6, 2,   3)
        __sampling(bins == 7, 3, np.inf)
        return sampled_results

class GANBLRPP:
    """
    The GANBLR++ model.

    Parameters
    ----------
    numerical_columns : list of int
        Indicating the indexes of numerical columns. 
        For example, if the 3, 5, 10th feature of a data is numerical feature, then this param should be [3, 5, 10].
    
    random_state : int, RandomState instance or None
        Controls the random seed given to the method chosen to initialize the parameters of `BayesianGaussianMixture` used by `GANBLRPP`.
    """
    def __init__(self, numerical_columns, random_state=None):
        self.__discritizer = DMMDiscritizer(random_state)
        self.__ganblr = GANBLR()
        self._numerical_columns = numerical_columns
        pass
    
    def fit(self, x, y, k=0, batch_size=32, epochs=10, warmup_epochs=1, verbose=1):
        '''
        Fit the model to the given data.

        Parameters
        ----------
        x : array_like of shape (n_samples, n_features)
            Dataset to fit the model. The data should be discrete.
        
        y : array_like of shape (n_samples,)
            Label of the dataset.

        k : int, default=0
            Parameter k of ganblr model. Must be greater than 0. No more than 2 is Suggested.

        batch_size : int, default=32
            Size of the batch to feed the model at each step.
        
        epochs : int, default=0
            Number of epochs to use during training.

        warmup_epochs : int, default=1
            Number of epochs to use in warmup phase. Defaults to :attr:`1`.
        
        verbose : int, default=1
            Whether to output the log. Use 1 for log output and 0 for complete silence.
        
        Returns
        -------
        self : object
            Fitted model.
        '''
        numerical_columns = self._numerical_columns
        x[:,numerical_columns] = self.__discritizer.fit_transform(x[:,numerical_columns])
        return self.__ganblr.fit(x, y, k, batch_size, epochs, warmup_epochs, verbose)
    
    def sample(self, size=None, verbose=1):
        """
        Generate synthetic data.     

        Parameters
        ----------
        size : int or None
            Size of the data to be generated. set to `None` to make the size equal to the size of the training set.

        verbose : int, default=1
            Whether to output the log. Use 1 for log output and 0 for complete silence.
        
        Return:
        -----------------
        synthetic_samples : np.ndarray
            Generated synthetic data.
        """
        if verbose:
            print('Step 1/2: Sampling discrete data from GANBLR.')
        ordinal_data = self.__ganblr._sample(size, verbose=verbose)
        syn_x = self.__ganblr._ordinal_encoder.inverse_transform(ordinal_data[:,:-1])
        syn_y = self.__ganblr._label_encoder.inverse_transform(ordinal_data[:,-1]).reshape(-1,1)
        if verbose:
            print('step 2/2: Sampling numerical data.')
        numerical_columns = self._numerical_columns
        numerical_data = self.__discritizer.inverse_transform(syn_x[:,numerical_columns].astype(int))
        syn_x[:,numerical_columns] = numerical_data 
        return np.hstack([syn_x, syn_y])

    def evaluate(self, x, y, model='lr'):
        """
        Perform a TSTR(Training on Synthetic data, Testing on Real data) evaluation.

        Parameters
        ----------
        x, y : array_like
            Test dataset.

        model : str or object
            The model used for evaluate. Should be one of ['lr', 'mlp', 'rf'], or a model class that have sklearn-style `fit` and `predict` method.

        Return:
        --------
        accuracy_score : float.

        """
        from sklearn.linear_model import LogisticRegression
        from sklearn.neural_network import MLPClassifier
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.preprocessing import OneHotEncoder, StandardScaler, OrdinalEncoder
        from sklearn.metrics import accuracy_score
    
        eval_model = None
        if model=='lr':
            eval_model = LogisticRegression() 
        elif model == 'rf':
            eval_model = RandomForestClassifier()
        elif model == 'mlp':
            eval_model = MLPClassifier() 
        elif hasattr(model, 'fit') and hasattr(model, 'predict'):
            eval_model = model
        else:
            raise Exception('Invalid Arugument')
        numerical_columns = self._numerical_columns
        catgorical_columns = list(set(range(x.shape[1])) - set(numerical_columns))  
        categories = self.__ganblr._d.get_categories(catgorical_columns)

        synthetic_data = self.sample()
        synthetic_x, synthetic_y = synthetic_data[:,:-1], synthetic_data[:,-1]
        
        ohe = OneHotEncoder(categories=categories, sparse=False, handle_unknown='ignore')
        syn_x_ohe  = ohe.fit_transform(synthetic_x[:,catgorical_columns])
        real_x_ohe = ohe.transform(x[:,catgorical_columns])
        syn_x_num  = synthetic_x[:,numerical_columns]
        real_x_num = x[:,numerical_columns]

        scaler = StandardScaler()        
        syn_x_concat  = scaler.fit_transform(np.hstack([syn_x_num, syn_x_ohe]))
        real_x_concat = scaler.transform(np.hstack([real_x_num, real_x_ohe]))     

        lbe = self.__ganblr._label_encoder
        real_y = lbe.transform(y)
        syn_y  = lbe.transform(synthetic_y)

        eval_model.fit(syn_x_concat, syn_y)
        pred = eval_model.predict(real_x_concat)
        return accuracy_score(real_y, pred)



import numpy as np
#import networkx as nx
from pyitlib import discrete_random_variable as drv

def build_graph(X, y, k=2):
  '''
  kDB algorithm

  Param:
  ----------------------
    
  Return:
  ----------------------
  graph edges
  '''
  #ensure data
  num_features = X.shape[1]
  x_nodes = list(range(num_features))
  y_node  = num_features

  #util func
  _x = lambda i:X[:,i]
  _x2comb = lambda i,j:(X[:,i], X[:,j])

  #feature indexes desc sort by mutual information
  sorted_feature_idxs = np.argsort([
    drv.information_mutual(_x(i), y) 
    for i in range(num_features)
  ])[::-1]

  #start building graph
  edges = []
  for iter, target_idx in enumerate(sorted_feature_idxs):
    target_node = x_nodes[target_idx]
    edges.append((y_node, target_node))

    parent_candidate_idxs = sorted_feature_idxs[:iter]
    if iter <= k:
      for idx in parent_candidate_idxs:
        edges.append((x_nodes[idx], target_node))
    else:
      first_k_parent_mi_idxs = np.argsort([
        drv.information_mutual_conditional(*_x2comb(i, target_idx), y)
        for i in parent_candidate_idxs
      ])[::-1][:k]
      first_k_parent_idxs = parent_candidate_idxs[first_k_parent_mi_idxs]

      for parent_idx in first_k_parent_idxs:
        edges.append((x_nodes[parent_idx], target_node))
  return edges

# def draw_graph(edges):
#   '''
#   Draw the graph
# 
#   Param
#   -----------------
#   edges: edges of the graph
# 
#   '''
#   graph = nx.DiGraph(edges)
#   pos=nx.spiral_layout(graph)
#   nx.draw(graph, pos, node_color='r', edge_color='b')
#   nx.draw_networkx_labels(graph, pos, font_size=20, font_family="sans-serif")


def get_cross_table(*cols, apply_wt=False):
    '''   
    author: alexland

    returns:
      (i) xt, NumPy array storing the xtab results, number of dimensions is equal to 
          the len(args) passed in
      (ii) unique_vals_all_cols, a tuple of 1D NumPy array for each dimension 
          in xt (for a 2D xtab, the tuple comprises the row and column headers)
      pass in:
        (i) 1 or more 1D NumPy arrays of integers
        (ii) if wts is True, then the last array in cols is an array of weights
        
    if return_inverse=True, then np.unique also returns an integer index 
    (from 0, & of same len as array passed in) such that, uniq_vals[idx] gives the original array passed in
    higher dimensional cross tabulations are supported (eg, 2D & 3D)
    cross tabulation on two variables (columns):
    >>> q1 = np.array([7, 8, 8, 8, 5, 6, 4, 6, 6, 8, 4, 6, 6, 6, 6, 8, 8, 5, 8, 6])
    >>> q2 = np.array([6, 4, 6, 4, 8, 8, 4, 8, 7, 4, 4, 8, 8, 7, 5, 4, 8, 4, 4, 4])
    >>> uv, xt = xtab(q1, q2)
    >>> uv
      (array([4, 5, 6, 7, 8]), array([4, 5, 6, 7, 8]))
    >>> xt
      array([[2, 0, 0, 0, 0],
             [1, 0, 0, 0, 1],
             [1, 1, 0, 2, 4],
             [0, 0, 1, 0, 0],
             [5, 0, 1, 0, 1]], dtype=uint64)
      '''
    if not all(len(col) == len(cols[0]) for col in cols[1:]):
      raise ValueError("all arguments must be same size")

    if len(cols) == 0:
      raise TypeError("xtab() requires at least one argument")

    fnx1 = lambda q: len(q.squeeze().shape)
    if not all([fnx1(col) == 1 for col in cols]):
      raise ValueError("all input arrays must be 1D")

    if apply_wt:
      cols, wt = cols[:-1], cols[-1]
    else:
      wt = 1

    uniq_vals_all_cols, idx = zip( *(np.unique(col, return_inverse=True) for col in cols) )
    shape_xt = [uniq_vals_col.size for uniq_vals_col in uniq_vals_all_cols]
    dtype_xt = 'float' if apply_wt else 'uint'
    xt = np.zeros(shape_xt, dtype=dtype_xt)
    np.add.at(xt, idx, wt)
    return uniq_vals_all_cols, xt

def _get_dependencies_without_y(variables, y_name, kdb_edges):
    ''' 
    evidences of each variable without y.

    Param:
    --------------
    variables: variable names

    y_name: class name

    kdb_edges: list of tuple (source, target)
    '''
    dependencies = {}
    kdb_edges_without_y = [edge for edge in kdb_edges if edge[0] != y_name]
    mi_desc_order = {t:i for i,(s,t) in enumerate(kdb_edges) if s == y_name}
    for x in variables:
        current_dependencies = [s for s,t in kdb_edges_without_y if t == x]
        if len(current_dependencies) >= 2:
            sort_dict = {t:mi_desc_order[t] for t in current_dependencies}        
            dependencies[x] = sorted(sort_dict)
        else:
            dependencies[x] = current_dependencies
    return dependencies

def _add_uniform(array, noise=1e-5):
    ''' 
    if no count on particular condition for any feature, give a uniform prob rather than leave 0
    '''
    sum_by_col = np.sum(array,axis=0)
    zero_idxs = (array == 0).astype(int)
    # zero_count_by_col = np.sum(zero_idxs,axis=0)
    nunique = array.shape[0]
    result = np.zeros_like(array, dtype='float')
    for i in range(array.shape[1]):
        if sum_by_col[i] == 0:
            result[:,i] = array[:,i] + 1./nunique
        elif noise != 0:
            result[:,i] = array[:,i] + noise * zero_idxs[:,i]
        else:
            result[:,i] = array[:,i]
    return result

def _normalize_by_column(array):
    sum_by_col = np.sum(array,axis=0)
    return np.divide(array, sum_by_col,
        out=np.zeros_like(array,dtype='float'),
        where=sum_by_col !=0)

def _smoothing(cct, d):
    '''
    probability smoothing for kdb
    
    Parameters:
    -----------
    cct (np.ndarray): cross count table with shape (x0, *parents)

    d (int): dimension of cct

    Return:
    --------
    smoothed joint prob table
    '''
    #covert cross-count-table to joint-prob-table by doing a normalization alone axis 0
    jpt = _normalize_by_column(cct)
    smoothing_idx = jpt == 0
    if d > 1 and np.sum(smoothing_idx) > 0:
        parent = cct.sum(axis=-1)
        parent = _smoothing(parent, d-1)
        parent_extend = parent.repeat(jpt.shape[-1]).reshape(jpt.shape)
        jpt[smoothing_idx] = parent_extend[smoothing_idx]
    return jpt

def get_high_order_feature(X, col, evidence_cols, feature_uniques):
    '''
    encode the high order feature of X[col] given evidences X[evidence_cols].
    '''
    if evidence_cols is None or len(evidence_cols) == 0:
        return X[:,[col]]
    else:
        evidences = [X[:,_col] for _col in evidence_cols]

        #[1, variable_unique, evidence_unique]
        base = [1, feature_uniques[col]] + [feature_uniques[_col] for _col in evidence_cols[::-1][:-1]]
        cum_base = np.cumprod(base)[::-1]
        
        cols = evidence_cols + [col]
        high_order_feature = np.sum(X[:,cols] * cum_base, axis=1).reshape(-1,1)
        return high_order_feature

def get_high_order_constraints(X, col, evidence_cols, feature_uniques):
    '''
    find the constraints infomation for the high order feature X[col] given evidences X[evidence_cols].
    
    Returns:
    ---------------------
    tuple(have_value, high_order_uniques)

    have_value: a k+1 dimensions numpy ndarray of type boolean. 
        Each dimension correspond to a variable, with the order (*evidence_cols, col)
        True indicate the corresponding combination of variable values cound be found in the dataset.
        False indicate not.

    high_order_constraints: a 1d nummy ndarray of type int.
        Each number `c` indicate that there are `c` cols shound be applying the constraints since the last constrant position(or index 0),
        in sequence.         

    '''
    if evidence_cols is None or len(evidence_cols) == 0:
        unique = feature_uniques[col]
        return np.ones(unique,dtype=bool), np.array([unique])
    else:
        cols = evidence_cols + [col]
        cross_table_idxs, cross_table = get_cross_table(*[X[:,i] for i in cols])
        have_value = cross_table != 0
    
        have_value_reshape = have_value.reshape(-1,have_value.shape[-1])
        #have_value_split = np.split(have_value_reshape, have_value_reshape.shape[0], 0)
        high_order_constraints = np.sum(have_value_reshape, axis=-1)
    
        return have_value, high_order_constraints

class KdbHighOrderFeatureEncoder:
    '''
    High order feature encoder that uses the kdb model to retrieve the dependencies between features.
    
    '''
    def __init__(self):
        self.dependencies_ = {}
        self.constraints_ = np.array([])
        self.have_value_idxs_ = []
        self.feature_uniques_ = []
        self.high_order_feature_uniques_ = []
        self.edges_ = []
        self.ohe_ = None
        self.k = None
        #self.full_=True
    
    def fit(self, X, y, k=0):
        '''
        Fit the KdbHighOrderFeatureEncoder to X, y.

        Parameters
        ----------
        X : array_like of shape (n_samples, n_features)
            data to fit in the encoder.

        y : array_like of shape (n_samples,)
            label to fit in the encoder.

        k : int, default=0
            k value of the order of the high-order feature. k = 0 will lead to a OneHotEncoder.

        Returns
        -------
        self : object
            Fitted encoder.
        '''
        self.k = k
        edges = build_graph(X, y, k)
        #n_classes = len(np.unique(y))
        num_features = X.shape[1]

        if k > 0:
            dependencies = _get_dependencies_without_y(list(range(num_features)), num_features, edges)
        else:
            dependencies = {x:[] for x in range(num_features)}
        
        self.dependencies_ = dependencies
        self.feature_uniques_ = [len(np.unique(X[:,i])) for i in range(num_features)]
        self.edges_ = edges
        #self.full_ = full

        Xk, constraints, have_value_idxs = self.transform(X, return_constraints=True, use_ohe=False)

        from sklearn.preprocessing import OneHotEncoder
        self.ohe_ = OneHotEncoder().fit(Xk)
        self.high_order_feature_uniques_ = [len(c) for c in self.ohe_.categories_]
        self.constraints_ = constraints
        self.have_value_idxs_ = have_value_idxs
        return self
        
    def transform(self, X, return_constraints=False, use_ohe=True):
        """
        Transform X to the high-order features.

        Parameters
        ----------
        X : array_like of shape (n_samples, n_features)
            Data to fit in the encoder.
        
        return_constraints : bool, default=False
            Whether to return the constraint informations. 
        
        use_ohe : bool, default=True
            Whether to transform output to one-hot format.

        Returns
        -------
        X_out : ndarray of shape (n_samples, n_encoded_features)
            Transformed input.
        """
        Xk = []
        have_value_idxs = []
        constraints = []
        for k, v in self.dependencies_.items():
            xk = get_high_order_feature(X, k, v, self.feature_uniques_)
            Xk.append(xk)

            if return_constraints:
                idx, constraint = get_high_order_constraints(X, k, v, self.feature_uniques_)
                have_value_idxs.append(idx)
                constraints.append(constraint)
        
        Xk = np.hstack(Xk)
        from sklearn.preprocessing import OrdinalEncoder
        Xk = OrdinalEncoder().fit_transform(Xk)
        if use_ohe:
            Xk = self.ohe_.transform(Xk)

        if return_constraints:
            concated_constraints = np.hstack(constraints)
            return Xk, concated_constraints, have_value_idxs
        else:
            return Xk
    
    def fit_transform(self, X, y, k=0, return_constraints=False):
        '''
        Fit KdbHighOrderFeatureEncoder to X, y, then transform X.
        
        Equivalent to fit(X, y, k).transform(X, return_constraints) but more convenient.

        Parameters
        ----------
        X : array_like of shape (n_samples, n_features)
            data to fit in the encoder.

        y : array_like of shape (n_samples,)
            label to fit in the encoder.

        k : int, default=0
            k value of the kdb model. k = 0 will lead to a OneHotEncoder.
        
        return_constraints : bool, default=False
            whether to return the constraint informations. 

        Returns
        -------
        X_out : ndarray of shape (n_samples, n_encoded_features)
            Transformed input.
        '''
        return self.fit(X, y, k).transform(X, return_constraints)


from ..kdb import *
from ..kdb import _add_uniform
from ..utils import *
from pgmpy.models import BayesianNetwork
from pgmpy.sampling import BayesianModelSampling
from pgmpy.factors.discrete import TabularCPD
from sklearn.preprocessing import OrdinalEncoder, LabelEncoder
import numpy as np
import tensorflow as tf

class GANBLR:
    """
    The GANBLR Model.
    """
    def __init__(self) -> None:
        self._d = None
        self.__gen_weights = None
        self.batch_size = None
        self.epochs = None
        self.k = None
        self.constraints = None
        self._ordinal_encoder = OrdinalEncoder(dtype=int, handle_unknown='use_encoded_value', unknown_value=-1)
        self._label_encoder   = LabelEncoder()
    
    def fit(self, x, y, k=0, batch_size=32, epochs=10, warmup_epochs=1, verbose=1):
        '''
        Fit the model to the given data.

        Parameters
        ----------
        x : array_like of shape (n_samples, n_features)
            Dataset to fit the model. The data should be discrete.
        
        y : array_like of shape (n_samples,)
            Label of the dataset.

        k : int, default=0
            Parameter k of ganblr model. Must be greater than 0. No more than 2 is Suggested.

        batch_size : int, default=32
            Size of the batch to feed the model at each step.
        
        epochs : int, default=0
            Number of epochs to use during training.

        warmup_epochs : int, default=1
            Number of epochs to use in warmup phase. Defaults to :attr:`1`.
        
        verbose : int, default=1
            Whether to output the log. Use 1 for log output and 0 for complete silence.
        
        Returns
        -------
        self : object
            Fitted model.
        '''
        if verbose is None or not isinstance(verbose, int):
            verbose = 1
        x = self._ordinal_encoder.fit_transform(x)
        y = self._label_encoder.fit_transform(y).astype(int)
        d = DataUtils(x, y)
        self._d = d
        self.k = k
        self.batch_size = batch_size
        if verbose:
            print(f"warmup run:")
        history = self._warmup_run(warmup_epochs, verbose=verbose)
        syn_data = self._sample(verbose=0)
        discriminator_label = np.hstack([np.ones(d.data_size), np.zeros(d.data_size)])
        for i in range(epochs):
            discriminator_input = np.vstack([x, syn_data[:,:-1]])
            disc_input, disc_label = sample(discriminator_input, discriminator_label, frac=0.8)
            disc = self._discrim()
            d_history = disc.fit(disc_input, disc_label, batch_size=batch_size, epochs=1, verbose=0).history
            prob_fake = disc.predict(x, verbose=0)
            ls = np.mean(-np.log(np.subtract(1, prob_fake)))
            g_history = self._run_generator(loss=ls).history
            syn_data = self._sample(verbose=0)
            
            if verbose:
                print(f"Epoch {i+1}/{epochs}: G_loss = {g_history['loss'][0]:.6f}, G_accuracy = {g_history['accuracy'][0]:.6f}, D_loss = {d_history['loss'][0]:.6f}, D_accuracy = {d_history['accuracy'][0]:.6f}")
        return self
        
    def evaluate(self, x, y, model='lr') -> float:
        """
        Perform a TSTR(Training on Synthetic data, Testing on Real data) evaluation.

        Parameters
        ----------
        x, y : array_like
            Test dataset.

        model : str or object
            The model used for evaluate. Should be one of ['lr', 'mlp', 'rf'], or a model class that have sklearn-style `fit` and `predict` method.

        Return:
        --------
        accuracy_score : float.

        """
        from sklearn.linear_model import LogisticRegression
        from sklearn.neural_network import MLPClassifier
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.preprocessing import OneHotEncoder
        from sklearn.pipeline import Pipeline
        from sklearn.metrics import accuracy_score
        
        eval_model = None
        models = dict(
            lr=LogisticRegression,
            rf=RandomForestClassifier,
            mlp=MLPClassifier
        )
        if model in models.keys():
            eval_model = models[model]()
        elif hasattr(model, 'fit') and hasattr(model, 'predict'):
            eval_model = model
        else:
            raise Exception("Invalid Arugument `model`, Should be one of ['lr', 'mlp', 'rf'], or a model class that have sklearn-style `fit` and `predict` method.")

        synthetic_data = self._sample()
        synthetic_x, synthetic_y = synthetic_data[:,:-1], synthetic_data[:,-1]
        x_test = self._ordinal_encoder.transform(x)
        y_test = self._label_encoder.transform(y)

        categories = self._d.get_categories()
        pipline = Pipeline([('encoder', OneHotEncoder(categories=categories, handle_unknown='ignore')), ('model',  eval_model)]) 
        pipline.fit(synthetic_x, synthetic_y)
        pred = pipline.predict(x_test)
        return accuracy_score(y_test, pred)
    
    def sample(self, size=None, verbose=1) -> np.ndarray:
        """
        Generate synthetic data.     

        Parameters
        ----------
        size : int or None
            Size of the data to be generated. set to `None` to make the size equal to the size of the training set.

        verbose : int, default=1
            Whether to output the log. Use 1 for log output and 0 for complete silence.
        
        Return:
        -----------------
        synthetic_samples : np.ndarray
            Generated synthetic data.
        """
        ordinal_data = self._sample(size, verbose)
        origin_x = self._ordinal_encoder.inverse_transform(ordinal_data[:,:-1])
        origin_y = self._label_encoder.inverse_transform(ordinal_data[:,-1]).reshape(-1,1)
        return np.hstack([origin_x, origin_y])
        
    def _sample(self, size=None, verbose=1) -> np.ndarray:
        """
        Generate synthetic data in ordinal encoding format
        """
        if verbose is None or not isinstance(verbose, int):
            verbose = 1
        #basic varibles
        d = self._d
        feature_cards = np.array(d.feature_uniques)
        #ensure sum of each constraint group equals to 1, then re concat the probs
        _idxs = np.cumsum([0] + d._kdbe.constraints_.tolist())
        constraint_idxs = [(_idxs[i],_idxs[i+1]) for i in range(len(_idxs)-1)]
        
        probs = np.exp(self.__gen_weights[0])
        cpd_probs = [probs[start:end,:] for start, end in constraint_idxs]
        cpd_probs = np.vstack([p/p.sum(axis=0) for p in cpd_probs])
    
        #assign the probs to the full cpd tables
        idxs = np.cumsum([0] + d._kdbe.high_order_feature_uniques_)
        feature_idxs = [(idxs[i],idxs[i+1]) for i in range(len(idxs)-1)]
        have_value_idxs = d._kdbe.have_value_idxs_
        full_cpd_probs = [] 
        for have_value, (start, end) in zip(have_value_idxs, feature_idxs):
            #(n_high_order_feature_uniques, n_classes)
            cpd_prob_ = cpd_probs[start:end,:]
            #(n_all_combination) Note: the order is (*parent, variable)
            have_value_ravel = have_value.ravel()
            #(n_classes * n_all_combination)
            have_value_ravel_repeat = np.hstack([have_value_ravel] * d.num_classes)
            #(n_classes * n_all_combination) <- (n_classes * n_high_order_feature_uniques)
            full_cpd_prob_ravel = np.zeros_like(have_value_ravel_repeat, dtype=float)
            full_cpd_prob_ravel[have_value_ravel_repeat] = cpd_prob_.T.ravel()
            #(n_classes * n_parent_combinations, n_variable_unique)
            full_cpd_prob = full_cpd_prob_ravel.reshape(-1, have_value.shape[-1]).T
            full_cpd_prob = _add_uniform(full_cpd_prob, noise=0)
            full_cpd_probs.append(full_cpd_prob)
    
        #prepare node and edge names
        node_names = [str(i) for i in range(d.num_features + 1)]
        edge_names = [(str(i), str(j)) for i,j in d._kdbe.edges_]
        y_name = node_names[-1]
    
        #create TabularCPD objects
        evidences = d._kdbe.dependencies_
        feature_cpds = [
            TabularCPD(str(name), feature_cards[name], table, 
                       evidence=[y_name, *[str(e) for e in evidences]], 
                       evidence_card=[d.num_classes, *feature_cards[evidences].tolist()])
            for (name, evidences), table in zip(evidences.items(), full_cpd_probs)
        ]
        y_probs = (d.class_counts/d.data_size).reshape(-1,1)
        y_cpd = TabularCPD(y_name, d.num_classes, y_probs)
    
        #create kDB model, then sample the data
        model = BayesianNetwork(edge_names)
        model.add_cpds(y_cpd, *feature_cpds)
        sample_size = d.data_size if size is None else size
        result = BayesianModelSampling(model).forward_sample(size=sample_size, show_progress = verbose > 0)
        sorted_result = result[node_names].values
        
        return sorted_result
    
    def _warmup_run(self, epochs, verbose=None):
        d = self._d
        tf.keras.backend.clear_session()
        ohex = d.get_kdbe_x(self.k)
        self.constraints = softmax_weight(d.constraint_positions)
        elr = get_lr(ohex.shape[1], d.num_classes, self.constraints)
        history = elr.fit(ohex, d.y, batch_size=self.batch_size, epochs=epochs, verbose=verbose)
        self.__gen_weights = elr.get_weights()
        tf.keras.backend.clear_session()
        return history

    def _run_generator(self, loss):
        d = self._d
        ohex = d.get_kdbe_x(self.k)
        tf.keras.backend.clear_session()
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Dense(d.num_classes, input_dim=ohex.shape[1], activation='softmax',kernel_constraint=self.constraints))
        model.compile(loss=elr_loss(loss), optimizer='adam', metrics=['accuracy'])
        model.set_weights(self.__gen_weights)
        history = model.fit(ohex, d.y, batch_size=self.batch_size,epochs=1, verbose=0)
        self.__gen_weights = model.get_weights()
        tf.keras.backend.clear_session()
        return history
    
    def _discrim(self):
        model = tf.keras.Sequential()
        model.add(tf.keras.layers.Dense(1, input_dim=self._d.num_features, activation='sigmoid'))
        model.compile(loss='binary_crossentropy', optimizer='adam', metrics=['accuracy'])
        return model


from tensorflow.python.ops import math_ops
import numpy as np
import tensorflow as tf

class softmax_weight(tf.keras.constraints.Constraint):
    """Constrains weight tensors to be under softmax `."""
  
    def __init__(self,feature_uniques):
        if isinstance(feature_uniques, np.ndarray):
            idxs = math_ops.cumsum(np.hstack([np.array([0]),feature_uniques]))
        else:
            idxs = math_ops.cumsum([0] + feature_uniques)
        idxs = [i.numpy() for i in idxs]
        self.feature_idxs = [
            (idxs[i],idxs[i+1]) for i in range(len(idxs)-1)
        ]
  
    def __call__(self, w):     
        w_new = [
            math_ops.log(tf.nn.softmax(w[i:j,:], axis=0))
            for i,j in self.feature_idxs
        ]
        return tf.concat(w_new, 0)
  
    def get_config(self):
        return {'feature_idxs': self.feature_idxs}

def elr_loss(KL_LOSS):
  def loss(y_true, y_pred):
    return tf.keras.losses.sparse_categorical_crossentropy(y_true, y_pred)+ KL_LOSS
  return loss

def KL_loss(prob_fake):
    return np.mean(-np.log(np.subtract(1,prob_fake)))

def get_lr(input_dim, output_dim, constraint=None,KL_LOSS=0):
    model = tf.keras.Sequential()
    model.add(tf.keras.layers.Dense(output_dim, input_dim=input_dim, activation='softmax',kernel_constraint=constraint))
    model.compile(loss=elr_loss(KL_LOSS), optimizer='adam', metrics=['accuracy'])
    #log_elr = model.fit(*train_data, validation_data=test_data, batch_size=batch_size,epochs=epochs)
    return model 

def sample(*arrays, n=None, frac=None, random_state=None):
    '''
    generate sample random arrays from given arrays. The given arrays must be same size.
    
    Parameters:
    --------------
    *arrays: arrays to be sampled.

    n (int): Number of random samples to generate.

    frac: Float value between 0 and 1, Returns (float value * length of given arrays). frac cannot be used with n.

    random_state: int value or numpy.random.RandomState, optional. if set to a particular integer, will return same samples in every iteration.

    Return:
    --------------
    the sampled array(s). Passing in multiple arrays will result in the return of a tuple.

    '''
    random = np.random
    if isinstance(random_state, int):
        random = random.RandomState(random_state)
    elif isinstance(random_state, np.random.RandomState):
        random = random_state
    
    arr0 = arrays[0]
    original_size = len(arr0)
    if n == None and frac == None:
        raise Exception('You must specify one of frac or size.')
    if n == None:
        n = int(len(arr0) * frac)

    idxs = random.choice(original_size, n, replace=False)
    if len(arrays) > 1:
        sampled_arrays = []
        for arr in arrays:
            assert(len(arr) == original_size)
            sampled_arrays.append(arr[idxs])
        return tuple(sampled_arrays)
    else:
        return arr0[idxs]

DEMO_DATASETS = {
    'adult': {
        'link':'https://raw.githubusercontent.com/chriszhangpodo/discretizedata/main/adult-dm.csv',
        'params': {
            'dtype' : int
        }
    },
    'adult-raw':{
        'link':'https://drive.google.com/uc?export=download&id=1iA-_qIC1xKQJ4nL2ugX1_XJQf8__xOY0',
        'params': {}
    }
}

def get_demo_data(name='adult'):
    """
    Download demo dataset from internet.

    Parameters
    ----------
    name : str 
        Name of dataset. Should be one of ['adult', 'adult-raw'].

    Returns
    -------
    data : pandas.DataFrame
        the demo dataset.
    """
    assert(name in DEMO_DATASETS.keys())
    return read_csv(DEMO_DATASETS[name]['link'], **DEMO_DATASETS[name]['params'])

from .kdb import KdbHighOrderFeatureEncoder
from sklearn.preprocessing import OneHotEncoder
from pandas import read_csv
import numpy as np

class DataUtils:
    """
    useful data utils for the preparation before training.
    """
    def __init__(self, x, y):
        self.x = x
        self.y = y
        self.data_size = len(x)
        self.num_features = x.shape[1]

        yunique, ycounts = np.unique(y, return_counts=True)
        self.num_classes = len(yunique)
        self.class_counts = ycounts
        self.feature_uniques = [len(np.unique(x[:,i])) for i in range(self.num_features)]
        
        self.constraint_positions = None
        self._kdbe = None

        self.__kdbe_x = None

    def get_categories(self, idxs=None):
        if idxs != None:
            return [self._kdbe.ohe_.categories_[i] for i in idxs]
        return self._kdbe.ohe_.categories_

    def get_kdbe_x(self, k=0, dense_format=True) -> np.ndarray:
        if self.__kdbe_x is not None:
            return self.__kdbe_x
        if self._kdbe == None:
            self._kdbe = KdbHighOrderFeatureEncoder()
            self._kdbe.fit(self.x, self.y, k=k)
        kdbex = self._kdbe.transform(self.x)
        if dense_format:
            kdbex = kdbex.todense()
        self.__kdbe_x = kdbex
        self.constraint_positions = self._kdbe.constraints_
        return kdbex
    
    def clear(self):
        self._kdbe = None
        self.__kdbe_x = None


import os
from setuptools import setup, find_packages

path = os.path.abspath(os.path.dirname(__file__))

try:
  with open(os.path.join(path, 'README.md')) as f:
    long_description = f.read()
except Exception as e:
  long_description = "Ganblr Toolbox"

setup(
    name = "ganblr",
    version = "0.1.1",
    keywords = ["ganblr", "tulip"],
    description = "Ganblr Toolbox",
    long_description = long_description,
    long_description_content_type='text/markdown',
    python_requires=">=3.5.0",
    license = "MIT Licence",

    url = "https://github.com/tulip-lab/ganblr",
    author = "kae zhou",
    author_email = "kaezhou@gmail.com",

    packages = find_packages(),
    include_package_data = True,
    install_requires = ["numpy", "pandas", "tensorflow>=2.3", "scikit-learn>=0.24", "pyitlib", "pgmpy"],
    platforms = "any",

    scripts = [],
    entry_points = {},

    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)